import datetime
import logging

LOG_FORMAT_KEY_VALUE = (
    "level=%(levelname)s "
    "ts=%(asctime)s "
    "process=%(processName)s "
    "process_id=%(process)s "
    "caller=%(name)s:%(lineno)s "
    "msg=\"%(message)s\""
)

LOG_FORMAT_DATE = "%Y-%m-%dT%H:%M:%S,%f"


class DateTimeFormatter(logging.Formatter):

    def formatTime(self, record, datefmt=None):
        converted_time = datetime.datetime.fromtimestamp(record.created)
        if datefmt:
            return converted_time.strftime(datefmt)
        formatted_time = converted_time.strftime("%Y-%m-%dT%H:%M:%S")
        return f"{formatted_time}.{record.msecs:03.0f}"


def setup_logging(filename: str):
    file_formatter = DateTimeFormatter(fmt=LOG_FORMAT_KEY_VALUE, datefmt=LOG_FORMAT_DATE)
    file_handler = logging.FileHandler(filename=filename)
    file_handler.formatter = file_formatter
    file_handler.level = logging.DEBUG
    try:
        # The first default handler is the StreamHandler <stderr>
        logging.getLogger().handlers[0].level = logging.ERROR
    except IndexError:
        pass
    logging.getLogger().addHandler(file_handler)
    logging.getLogger().level = logging.DEBUG



from rich.align import Align
from rich.columns import Columns
from rich.padding import Padding
from textual.reactive import Reactive
from textual.widgets import Footer


class Footer(Footer):

    log_size = Reactive(0)
    log_offset = Reactive(0)

    def on_mount(self) -> None:
        self.layout_size = 1

    def render(self) -> Columns:
        log_size_text = Align.right(
            Padding(
                f"at {self.log_offset} in [bold]{self.log_size}[/] lines", pad=(0, 1, 0, 1),
                style="white on dark_green",
                expand=False,
            )
        )

        return Columns([super().render(), log_size_text], expand=True)



import functools
import logging
import time

MODULE_LOGGER = logging.getLogger("Textual")


def timer(*, level: int = logging.INFO, precision: int = 4):
    """
    Print the runtime of the decorated function.

    Args:
        level: the logging level for the time message [default=INFO]
        precision: the number of decimals for the time [default=3 (ms)]
    """

    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper_timer(*args, **kwargs):
            start_time = time.perf_counter()
            value = func(*args, **kwargs)
            end_time = time.perf_counter()
            run_time = end_time - start_time
            MODULE_LOGGER.log(level, f"Finished {func.__name__!r} in {run_time:.{precision}f} secs")
            return value

        return wrapper_timer
    return actual_decorator


def do_every(period: float, func: callable, *args) -> None:
    """

    This method executes a function periodically, taking into account
    that the function that is executed will take time also and using a
    simple `sleep()` will cause a drift. This method will not drift.

    You can use this function in combination with the threading module to execute the
    function in the background, but be careful as the function might not be thread safe.

    ```
    timer_thread = threading.Thread(target=do_every, args=(10, func))
    timer_thread.daemon = True
    timer_thread.start()
    ```

    Args:
        period: a time interval between successive executions [seconds]
        func: the function to be executed
        *args: optional arguments to be passed to the function
    """

    # Code from SO:https://stackoverflow.com/a/28034554/4609203
    # The max in the yield line serves to protect sleep from negative numbers in case the
    # function being called takes longer than the period specified. In that case it would
    # execute immediately and make up the lost time in the timing of the next execution.

    def g_tick():
        next_time = time.time()
        while True:
            next_time += period
            yield max(next_time - time.time(), 0)

    g = g_tick()
    while True:
        time.sleep(next(g))
        func(*args)



from rich.panel import Panel
from textual.widget import Widget

from textualog import styles
from textualog.emojis import INFO
from textualog.renderables.shortcuts import Shortcuts


class Help(Widget):
    def render(self) -> Panel:
        return Panel(
            Shortcuts(),
            title=f"{INFO} [bold]help[/]",
            border_style=styles.BORDER_FOCUSED,
            box=styles.BOX,
            title_align="left",
            padding=0,
        )



import logging

from rich.panel import Panel
from rich.table import Table
from textual.reactive import Reactive
from textual.widget import Widget

from textualog import styles
from textualog.emojis import CHECK
from textualog.emojis import UNCHECK

PANEL_SIZE = 5


class Levels(Widget):

    debug_level: Reactive = Reactive(True)
    info_level: Reactive = Reactive(True)
    warning_level: Reactive = Reactive(True)
    error_level: Reactive = Reactive(True)
    critical_level: Reactive = Reactive(True)

    async def on_mount(self) -> None:
        self.layout_size = PANEL_SIZE

    def render(self) -> Panel:
        table = Table(box=None, expand=False, show_header=False, show_edge=False)
        table.add_column()
        table.add_column()

        table.add_row(CHECK if self.debug_level else UNCHECK, "DEBUG")
        table.add_row(CHECK if self.info_level else UNCHECK, "INFO")
        table.add_row(CHECK if self.warning_level else UNCHECK, "WARNING")
        table.add_row(CHECK if self.error_level else UNCHECK, "ERROR")
        table.add_row(CHECK if self.critical_level else UNCHECK, "CRITICAL")

        panel = Panel(
            table,
            title="[bold]Levels[/]",
            border_style=styles.BORDER,
            box=styles.BOX,
            title_align="left",
            padding=0,
        )

        return panel

    def is_on(self, level: int):
        if level == logging.DEBUG:
            return self.debug_level
        elif level == logging.INFO:
            return self.info_level
        elif level == logging.WARNING:
            return self.warning_level
        elif level == logging.ERROR:
            return self.error_level
        elif level == logging.CRITICAL:
            return self.critical_level
        return False

    def __str__(self):
        return (
            f"DEBUG={'ON' if self.debug_level else 'OFF'}, "
            f"INFO={'ON' if self.info_level else 'OFF'}, "
            f"WARNING={'ON' if self.warning_level else 'OFF'}, "
            f"ERROR={'ON' if self.error_level else 'OFF'}, "
            f"CRITICAL={'ON' if self.critical_level else 'OFF'}"
        )



import datetime
import logging
import time
from enum import Enum
from typing import Union

import rich
from rich.text import Text


class LevelColor(Enum):
    DEBUG = "white"
    INFO = "green"
    WARNING = "orange3"
    ERROR = "red"
    CRITICAL = "bright_magenta"


class LevelColorSelected(Enum):
    DEBUG = "grey93 on grey46"
    INFO = "green3 on dark_green"
    WARNING = "orange1 on dark_orange3"
    ERROR = "orchid2 on red3"
    CRITICAL = "magenta1 on dark_magenta"


class LevelName(Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class LogRecord:
    def __init__(
            self,
            msg: str,
            level: int = logging.INFO,
            created: float = None,
            ts: str = None,
            process: str = None,
            caller: str = None,
            process_id: int = None,
            selected: bool = False,
            extra: str = None,
            **kwargs,
    ):
        self.msg = msg
        self.level = level
        self.ts = ts
        self.created = created or to_timestamp(self.ts) if self.ts else time.time()
        self.process = process
        self.caller = caller
        self.process_id = process_id
        self.selected = selected
        self.extra = extra

    def __str__(self) -> str:
        text = (
            f"level={self.level} "
            f"ts={format_datetime(datetime.datetime.fromtimestamp(self.created))} "
            f"msg=\"{self.msg}\""
        )
        return text

    def get_text(self) -> Text:
        level = LevelName(self.level).name
        color = LevelColorSelected[level].value if self.selected else LevelColor[level].value
        return Text(
            f"{format_datetime(from_timestamp(self.created))} "
            f"{LevelName(self.level).name} "
            f"{self.msg}",
            style=f"{color}"
        )

    def __rich__(self) -> Text:
        color = LevelColor[LevelName(self.level).name].value
        return Text(
            f"{format_datetime(from_timestamp(self.created))} "
            f"{LevelName(self.level).name} "
            f"{self.msg}",
            style=f"{color}"
        )


def format_datetime(dt: Union[str, datetime.datetime] = None, fmt: str = None, width: int = 6, precision: int = 3):
    """Format a datetime as YYYY-mm-ddTHH:MM:SS.μs+0000.

    If the given argument is not timezone aware, the last part, i.e. `+0000` will not be there.

    If no argument is given, the timestamp is generated as
    `datetime.datetime.now(tz=datetime.timezone.utc)`.

    The `dt` argument can also be a string with the following values: today, yesterday, tomorrow,
    and 'day before yesterday'. The format will then be '%Y%m%d' unless specified.

    Optionally, a format string can be passed in to customize the formatting of the timestamp.
    This format string will be used with the `strftime()` method and should obey those conventions.

    Example:
        >>> format_datetime(datetime.datetime(2020, 6, 13, 14, 45, 45, 696138))
        '2020-06-13T14:45:45.696'
        >>> format_datetime(datetime.datetime(2020, 6, 13, 14, 45, 45, 696138), precision=6)
        '2020-06-13T14:45:45.696138'
        >>> format_datetime(datetime.datetime(2020, 6, 13, 14, 45, 59, 999501), precision=3)
        '2020-06-13T14:45:59.999'
        >>> format_datetime(datetime.datetime(2020, 6, 13, 14, 45, 59, 999501), precision=6)
        '2020-06-13T14:45:59.999501'
        >>> _ = format_datetime()
        ...
        >>> format_datetime("yesterday")
        '20220214'
        >>> format_datetime("yesterday", fmt="%d/%m/%Y")
        '14/02/2022'

    Args:
        dt (datetime): a datetime object or an agreed string like yesterday, tomorrow, ...
        fmt (str): a format string that is accepted by `strftime()`
        width (int): the width to use for formatting the microseconds
        precision (int): the precision for the microseconds
    Returns:
        a string representation of the current time in UTC, e.g. `2020-04-29T12:30:04.862+0000`.

    Raises:
        A ValueError will be raised when the given dt argument string is not understood.
    """
    dt = dt or datetime.datetime.now(tz=datetime.timezone.utc)
    if isinstance(dt, str):
        fmt = fmt or "%Y%m%d"
        if dt.lower() == "yesterday":
            dt = datetime.date.today() - datetime.timedelta(days=1)
        elif dt.lower() == "today":
            dt = datetime.date.today()
        elif dt.lower() == "day before yesterday":
            dt = datetime.date.today() - datetime.timedelta(days=2)
        elif dt.lower() == "tomorrow":
            dt = datetime.date.today() + datetime.timedelta(days=1)
        else:
            raise ValueError(f"Unknown date passed as an argument: {dt}")

    if fmt:
        timestamp = dt.strftime(fmt)
    else:
        width = min(width, precision)
        timestamp = (
            f"{dt.strftime('%Y-%m-%dT%H:%M')}:"
            f"{dt.second:02d}.{dt.microsecond//10**(6-precision):0{width}d}{dt.strftime('%z')}"
        )

    return timestamp


def from_timestamp(ts: float):
    return datetime.datetime.fromtimestamp(ts)


def to_timestamp(ts: str):
    return datetime.datetime.strptime(ts, '%Y-%m-%dT%H:%M:%S,%f').timestamp()


if __name__ == "__main__":

    for level in LevelName:
        record = LogRecord(level=level.value,
                           msg=f"A simple {level.name} message should be printed in {LevelColor[level.name].value}.")
        print(record)
        rich.print(record)



import argparse
import logging
import sys
import threading
from pathlib import Path

from textual import events
from textual.app import App
from textual.keys import Keys
from textual.reactive import Reactive
from textual.widgets import Header
from textual.widgets import ScrollView

from . import __version__
from .loader import KeyValueLoader
from .log import setup_logging
from .renderables.namespace_tree import EntryClick
from .system import do_every
from .system import timer
from .widgets.details import Details
from .widgets.footer import Footer
from .widgets.help import Help
from .widgets.levels import Levels
from .widgets.namespaces import Namespaces
from .widgets.recordinfo import RecordInfo
from .widgets.records import Records

logging.basicConfig(level=logging.ERROR)

MODULE_LOGGER = logging.getLogger("Textual")


class TextualLog(App):

    show_help = Reactive(False)
    show_namespaces = Reactive(False)
    show_details = Reactive(False)

    # The namespace_tree is just for demonstration purposes. The namespace should be a
    # tree like structure with proper navigation and the possibility to add and remove nodes.

    namespace_tree = {
        "egse": {
            "system": "system",
            "decorators": {"x": 1, "y": 2},
        }
    }

    def __init__(self, filename: str = None, **kwargs):
        super().__init__(**kwargs)
        self.filename = filename
        self.cursor = 0
        self.loader = None
        self.details_widget = None
        self.follow = False

    async def on_mount(self, event: events.Mount) -> None:
        """
        Call after terminal goes in to application mode.
        """

        self.namespaces = Namespaces("Name space", self.namespace_tree)
        self.namespaces.layout_offset_x = -40

        self.help_widget = Help()
        self.help_widget.visible = False

        self.details_widget = Details()
        self.details_scroll_view = ScrollView(self.details_widget)
        self.details_scroll_view.visible = False

        self.header = Header()
        self.footer = Footer()

        await self.view.dock(self.header, edge="top")
        await self.view.dock(self.footer, edge="bottom")
        await self.view.dock(self.namespaces, edge="left", size=40, z=1)
        await self.view.dock(self.help_widget, edge="right", size=40, z=1)
        await self.view.dock(self.details_scroll_view, z=0)
        grid = await self.view.dock_grid(edge="left", name="left")

        grid.add_column(size=30, name="left")
        grid.add_column(fraction=1, name="right", min_size=100)

        grid.add_row(fraction=1, name="top")
        grid.add_row(fraction=1, name="middle")
        grid.add_row(size=7, name="bottom")

        grid.add_areas(
            area1="left-start|right-end,top-start|middle-end",
            area2="left,bottom",
            area3="right,bottom",
        )

        self.levels = Levels()
        self.records = Records()
        self.record_info = RecordInfo()

        grid.place(
            area1=self.records,
            area2=self.levels,
            area3=self.record_info,
        )

        if self.filename:
            self.loader = KeyValueLoader(self.filename)
            self.loader.load()
            self.loader.process(0, 500, None)
            self.footer.log_size = self.loader.size()

            # The height of the self.records view is not yet known, so we take a large enough number

            self.records.update(self.loader.get_records(0, 500, None))

        # self.set_interval(2.0, self.collect_data)

        self._reload_thread = threading.Thread(
            target=do_every, args=(2.0, self.collect_data))
        self._reload_thread.daemon = True
        self._reload_thread.start()

    @timer(level=logging.DEBUG)
    def collect_data(self):
        if not self.follow or self.loader is None:
            return

        self.loader.load()

        # The height of the text area of the Records panel

        height = self.records.size.height - 2
        size = self.loader.size()

        self.cursor = max(0, size - height)
        self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        self.records.refresh(layout=True)

        self.footer.log_size = size

    async def on_load(self) -> None:
        """
        Sent before going in to application mode. This method is called before on_mount().
        Bind keys here.
        """
        await self.bind("q", "quit", "Quit")
        await self.bind("?", "toggle_help", "Help")

    async def on_key(self, event) -> None:

        if self.loader is None:
            return

        # The height of the text area of the Records panel

        height = self.records.size.height - 2
        size = self.loader.size()

        self.app.sub_title = f"Key pressed: {event.key}"

        if event.key == "d":
            self.levels.debug_level = not self.levels.debug_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == "i":
            self.levels.info_level = not self.levels.info_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == "w":
            self.levels.warning_level = not self.levels.warning_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == "e":
            self.levels.error_level = not self.levels.error_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == "c":
            self.levels.critical_level = not self.levels.critical_level
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key in "nN":
            self.show_namespaces = not self.show_namespaces
        elif event.key == "r":
            self.loader.load()
            self.footer.log_size = self.loader.size()
        elif event.key == "f":
            self.follow = not self.follow
            self.header.style = "white on dark_red" if self.follow else "white on dark_green"
            # self.collect_data()
        elif event.key == Keys.Escape:
            self.show_help = False
            self.show_namespaces = False
            self.show_details = False
        elif event.key == Keys.Down:
            self.cursor = min(size, self.cursor + 1)
            self.footer.log_offset = self.cursor
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.Up:
            self.cursor = max(0, self.cursor - 1)
            self.footer.log_offset = self.cursor
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.PageDown:
            self.cursor = min(size, self.cursor+(height-1))
            self.footer.log_offset = self.cursor
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.PageUp:
            self.cursor = max(0, self.cursor-(height-1))
            self.footer.log_offset = self.cursor
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.End:
            self.cursor = max(0, size - height)
            self.footer.log_offset = self.cursor
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))
        elif event.key == Keys.Home:
            self.cursor = 0
            self.footer.log_offset = self.cursor
            self.records.replace(self.loader.get_records(self.cursor, height, self.levels))

        self.records.refresh(layout=True, repaint=True)

    async def watch_show_namespaces(self, show_namespaces: bool) -> None:
        """Called when show_namespaces changes."""
        self.namespaces.animate("layout_offset_x", 0 if show_namespaces else -40)

    async def watch_show_help(self, show_help: bool) -> None:
        """Called when show_help changes."""
        self.help_widget.visible = show_help

    async def watch_show_details(self, show_details: bool) -> None:
        """Called when show_details changes."""
        self.details_widget.refresh(layout=True)
        self.details_scroll_view.visible = show_details

    async def action_toggle_help(self) -> None:
        """Called when bound help key is pressed."""
        self.show_help = not self.show_help

    async def handle_entry_click(self, message: EntryClick) -> None:
        """A message sent by the namespace tree when an entry is clicked."""

        self.app.sub_title = f"{message.key}"
        self.records.refresh(layout=True)


def main():
    from rich.traceback import install
    install(show_locals=False)

    parser = argparse.ArgumentParser(
        description="Textual Log Viewer, display, filter and search log files",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=_get_version_text(),
        help="display version information",
    )

    parser.add_argument(
        "--log",
        "-l",
        type=str,
        default=None,
        help="the full path to the log file that you want to follow",
    )

    parser.add_argument(
        "--debug",
        "-d",
        action="store_true",
        default=False,
        help="send debugging information to the debug log files",
    )

    args = parser.parse_args()

    if args.log and not Path(args.log).exists():
        raise FileNotFoundError(f"No such file {args.log}")

    log_filename = "textual.log" if args.debug else None

    if args.debug:
        setup_logging("textualog.log")

    TextualLog.run(title="Textual Log Viewer", log=log_filename, filename=args.log)


def _get_version_text():
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"

    return "\n".join(
        [
            f"textualog {__version__.__version__} [Python {python_version}]",
            "Copyright © 2022 Rik Huygen",
        ]
    )


if __name__ == "__main__":
    main()



from typing import Optional

from rich.panel import Panel
from rich.text import Text
from textual.widget import Widget
from textual.reactive import Reactive

from textualog import styles
from textualog.renderables.logrecord import LogRecord


class Details(Widget):

    record: Reactive[LogRecord] = Reactive(None)

    def __init__(self):
        super().__init__()
        self.record: Optional[LogRecord] = None

    def set(self, record: LogRecord):
        self.record = record

    def render(self) -> Panel:
        return Panel(
            self._generate_renderable(),
            title="[bold]Record Details[/]",
            border_style=styles.BORDER_FOCUSED,
            box=styles.BOX,
            title_align="left",
            padding=0,
        )

    def _generate_renderable(self) -> Text:

        if self.record is None:
            return Text()

        record = Text(no_wrap=True)
        record.append(f"level      = {self.record.level}\n")
        record.append(f"process    = {self.record.process}\n")
        record.append(f"process ID = {self.record.process_id}\n")
        record.append(f"caller     = {self.record.caller}\n")
        record.append(f"msg        = {self.record.msg}\n")
        record.append(f"extra      = {self.record.extra}\n")

        return record



APPROXIMATION = "\u2248"
UP = "\u2191"
DOWN = "\u2193"
LEFT = "\u2190"
RIGHT = "\u2192"
HOME = "\u2196"
END = "\u2198"
PAGE_DOWN = "\u21A7"
PAGE_UP = "\u21A5"
RIGHT_TRIANGLE = "\u25B6"
BIG_RIGHT_TRIANGLE = "\uE0B0"
DOWN_TRIANGLE = "\u25BC"
DIAMOND = "\u25C8"



from __future__ import annotations

from typing import Optional

from rich.panel import Panel
from rich.text import Text
from textual import events
from textual.reactive import Reactive
from textual.widget import Widget

from .. import styles
from ..renderables.logrecord import LogRecord

PANEL_SIZE = 5


class RecordInfo(Widget):

    record: Reactive[LogRecord] = Reactive(None)

    def __init__(self, height: int | None = None):
        super().__init__()
        self.record: Optional[LogRecord] = None

    async def on_mount(self) -> None:
        # self.layout_size = PANEL_SIZE
        # self.layout_fraction = 1
        ...

    async def on_click(self, event: events.Click) -> None:
        self.app.show_details = True

    def set(self, record: LogRecord):
        self.record = record
        self.app.details_widget.set(self.record)

    def render(self) -> Panel:

        return Panel(
            self._generate_renderable(),
            title=f"[bold]Record Info"
                  f"{' *' if self.record is not None and self.record.extra else ''}"
                  f"[/]",
            border_style=styles.BORDER_FOCUSED,
            box=styles.BOX,
            title_align="left",
            padding=0,
            # height=PANEL_SIZE,
        )

    def _generate_renderable(self) -> Text:

        if self.record is None:
            return Text()

        record = Text(no_wrap=True)
        record.append(f"level      = {self.record.level}\n")
        record.append(f"process    = {self.record.process}\n")
        record.append(f"process ID = {self.record.process_id}\n")
        record.append(f"caller     = {self.record.caller}\n")
        record.append(f"msg        = {self.record.msg}\n")

        return record
